
/*
 * Required headers
 */
#include <unistd.h> /* pipe, fork, select, sleep, execl */
#include <sys/time.h> /* select, gettimeofday */
#include <sys/types.h> /* fork, select */
#include <stdlib.h> /* realloc */
#include <stdio.h> /* snprintf */

#include "headers.h" /* system-specific stuff */
#include "messages.h" /* ERROR, INFO */

/*
 * Prototypes
 */
#include "processus.h"

/*
 * Functions
 */
#ifdef TEST_SUITE

int main( int i_argc, char *ppsz_argv[] )
{
    int p_fd[ 2 ];
    char psz_port[ 10 ];
    char *psz_program = ppsz_argv[ 0 ];

    switch( i_argc )
    {
        case 1:

            /* Called with no arguments, we are the forking program */
            if( pipe( p_fd ) )
            {
                ERROR( "Failed creating pipe" );
                return -1;
            }
            INFO( "Created pipes" );

            switch( fork() )
            {
                /* We are the child, run a program which will write data
                 * to filedescriptor p_fd[ 1 ] every 5 seconds */
                case 0:
                    snprintf( psz_port, 10, "%i", p_fd[ 1 ] );
                    INFO( "I'm the child, running `%s %s'",
                          psz_program, psz_port );
                    execl( psz_program, psz_program, psz_port, NULL );
                    break;
    
                /* Boo. */
                case -1:
                    ERROR( "Fork failed" );
                    break;

                /* We are the parent. Just wait for data... */
                default:
                    INFO( "I'm the parent, listening to stdin"
                          " and filedescriptor %i", p_fd[ 0 ] );
                    WaitForData( p_fd[ 0 ] );
                    break;
            }
            break;

        /* We got called by the child of the above fork */
        case 2:
            INFO( "I'm the exevl()ed program, writing to"
                  " filedescriptor %i", atoi( ppsz_argv[ 1 ] ) );
            WriteData( atoi( ppsz_argv[ 1 ] ) );
            break;
    }

    /* Everything was OK */
    INFO( "Test successful" );
    return 0;
}
#endif

void WaitForData( int i_fd )
{
    /* Descriptor set and timer for select */
    fd_set fdset;
    struct timeval timer;

    /* Descriptor which got data */
    int i_listen = -1;

    /* Buffer to store read data */
    char *psz_buf = NULL;
    int i_buf_size = 0;
    int i_buf_pos = 0;

    do
    {
        /* Initialize descriptor set */
        FD_ZERO( &fdset );
        FD_SET( 0, &fdset );
        FD_SET( i_fd, &fdset );

        /* Initialize timer */
        timer.tv_sec = 0;
        timer.tv_usec = 10000;

        /* Wait for data */
        if( select( i_fd + 1, &fdset, NULL, NULL, &timer ) )
        {
            int i_read;

            /* Check which filedescriptor got the data */
            if( FD_ISSET( 0, &fdset ) )
            {
                i_listen = 0;
            }
            else if( FD_ISSET( i_fd, &fdset ) )
            {
                i_listen = i_fd;
            }
            else
            {
                continue;
            }

            /* Check that our buffer is large enough */
            if( i_buf_pos + ALLOC_ONCE > i_buf_size )
            {
                i_buf_size += ALLOC_ONCE;
                psz_buf = realloc( psz_buf, i_buf_size );
                if( psz_buf == NULL )
                {
                    ERROR( "WaitForData: failed reallocation" );
                    return;
                }
            }

            /* Read data */
            i_read = read( i_listen, psz_buf + i_buf_pos, READ_ONCE );
            if( i_read < 0 )
            {
                ERROR( "WaitForData: error while reading" );
                free( psz_buf );
                return;
            }
            i_buf_pos += i_read;
        }
        else
        {
            /* No data: print what we got last time */
            if( i_buf_pos != 0 )
            {
                psz_buf[ i_buf_pos ] = '\0';
                INFO( "Got %i bytes from %s: %s", i_buf_pos,
                      i_listen == 0 ? "keyboard" : "pipe", psz_buf );
                i_buf_pos = 0;
            }
        }
    }
    while( 1 );
}

void WriteData( int i_fd )
{
    char psz_date[ 128 ];
    struct timeval tv_date;

    do
    {
        /* Writing the microseconds timestamp to psz_date */
        gettimeofday( &tv_date, NULL );
        snprintf( psz_date, 128, "-=([%i.%.6i])=-",
                  (int)tv_date.tv_sec, (int)tv_date.tv_usec );

        /* Sending psz_date to the filedescriptor */
        write( i_fd, psz_date, strlen( psz_date ) );

        /* Wait a bit */
        sleep( 5 );
    }
    while( 1 );
}

