
/*
 * Required headers
 */
#include <sys/types.h> /* socket, fork */
#include <sys/time.h> /* select */
#include <sys/socket.h> /* socket, inet_ntoa */
#include <netinet/in.h> /* htons, htonl, inet_ntoa */
#include <arpa/inet.h> /* inet_ntoa */
#include <netdb.h> /* gethostbyname */
#include <unistd.h> /* fork */
#include <signal.h> /* signal */
#include <stdlib.h> /* realloc, free */
#include <string.h> /* strcmp */

#include "headers.h" /* system-specific stuff */
#include "messages.h" /* ERROR, INFO */

/*
 * Prototypes
 */
#include "sockets.h"
int  SocketListen( void );
int  SocketWrite( char *psz_server );
void SocketHandle( int i_fd );

#define DEFAULT_PORT 2000
#define MAX_CONNECT 20

/*
 * Functions
 */
#ifdef TEST_SUITE

int main( int i_argc, char *ppsz_argv[] )
{
    switch( i_argc )
    {
        /* No arguments, we are the listener */
        case 1:
            INFO( "Creating socket for listening" );
            return SocketListen();

        /* One argument, we are the talker */
        case 2:
            INFO( "Creating socket and connecting to %s", ppsz_argv[ 1 ] );
            return SocketWrite( ppsz_argv[ 1 ] );
    }

    /* Everything was OK */
    INFO( "Test successful" );
    return 0;
}
#endif

int SocketListen( void )
{
    int i_ret;
    int i_dummy;
    fd_set fdset;
    struct timeval timer;

    /* Server information */
    int i_fd;
    struct sockaddr_in server;

    /* Client information */
    int i_newfd;
    struct sockaddr_in client;
    socklen_t addrlen;

    /* Ignore SIG_CHLD to avoid zombies */
    signal( SIGCHLD, SIG_IGN );

    /* Create socket */
    i_fd = socket( PF_INET, SOCK_STREAM, 0 );
    if( i_fd < 0 )
    {
        ERROR( "Failed to create socket" );
        return -1;
    }

    /* Configure socket to reuse address */
    i_dummy = 1;
    i_ret = setsockopt( i_fd, SOL_SOCKET, SO_REUSEADDR, (void *) &i_dummy,
                        sizeof( i_dummy ) );
    if( i_ret == -1 )
    {
        ERROR( "Failed to setsockopt( SO_REUSEADDR )" );
        return -1;
    }

    /* Set remaining socket information */
    memset( &server, 0x00, sizeof( struct sockaddr_in ) );
    server.sin_family = AF_INET;
    server.sin_port = htons( DEFAULT_PORT );
    server.sin_addr.s_addr = htonl( INADDR_ANY );

    /* Bind socket */
    i_ret = bind( i_fd, (struct sockaddr *) &server, sizeof( server ) );
    if( i_ret < 0 )
    {
        ERROR( "Failed to bind socket" );
        return -1;
    }

    /* Set socket to listen mode */
    i_ret = listen( i_fd, MAX_CONNECT );
    if( i_ret < 0 )
    {
        ERROR( "Failed to listen to socket" );
        return -1;
    }

    /* Loop until we get a connection */
    do
    {
        FD_ZERO( &fdset );
        FD_SET( i_fd, &fdset );

        timer.tv_sec = 0;
        timer.tv_usec = 10000;

        /* Wait for a connection */
        if( select( i_fd + 1, &fdset, NULL, NULL, &timer ) )
        {
            i_newfd = accept( i_fd, (struct sockaddr *) &client, &addrlen );
            if( i_newfd == -1 )
            {
                ERROR( "Failed accepting connection" );
                continue;
            }

            INFO( "Got a connection from %s", inet_ntoa( client.sin_addr ) );

            switch( fork() )
            {
                /* We are the child, handle the socket */
                case 0:
                    INFO( "I'm a new child, handling fd %i", i_newfd );
                    SocketHandle( i_newfd );
                    return 0;
                    break;
    
                /* Boo. */
                case -1:
                    ERROR( "Fork failed" );
                    break;

                /* We are the parent. Just wait for data... */
                default:
                    if( close( i_newfd ) )
                      ERROR( "Parent failed to close connection %d", i_newfd );
                    break;
            }
        }

    } while( 1 );

    return 0;
}

int SocketWrite( char *psz_server )
{
    int i_ret;

    /* Socket information */
    int i_fd;
    struct sockaddr_in server;

    /* Create socket */
    i_fd = socket( AF_INET, SOCK_STREAM, 0 );
    if( i_fd < 0 )
    {
        ERROR( "Failed to create socket" );
        return -1;
    }

    /* Set remaining socket information */
    memset( &server, 0x00, sizeof( struct sockaddr_in ) );
    server.sin_family = AF_INET;
    server.sin_port = htons( DEFAULT_PORT );
    if(!inet_aton(psz_server, &server.sin_addr))
    {
        struct hostent* p_hostent = gethostbyname(psz_server);
        if(p_hostent == NULL)
        {
            ERROR( "Invalid server name %s", psz_server );
            return -1;
        }

        memcpy(&server.sin_addr, p_hostent->h_addr, p_hostent->h_length);
    }

    /* Bind socket */
    i_ret = connect( i_fd, (struct sockaddr *) &server, sizeof( server ) );
    if( i_ret < 0 )
    {
        ERROR( "Failed to connect socket" );
        return -1;
    }

    i_ret = write( i_fd, "blabla", 6);
    if( i_ret < 0 )
    {
        ERROR( "Failed to write data" );
        return -1;
    }

    if( close( i_fd ) )
    {
        ERROR( "Failed to close socket" );
        return -1;
    }

    return -0;
}

void SocketHandle( int i_fd )
{
    /* Descriptor set and timer for select */
    fd_set fdset;
    struct timeval timer;
    int i_alive = 1;

    /* Buffer to store read data */
    char *psz_buf = NULL;
    int i_buf_size = 0;
    int i_buf_pos = 0;

    do
    {
        /* Initialize descriptor set */
        FD_ZERO( &fdset );
        FD_SET( i_fd, &fdset );

        /* Initialize timer */
        timer.tv_sec = 0;
        timer.tv_usec = 10000;

        /* Wait for data */
        if( select( i_fd + 1, &fdset, NULL, NULL, &timer ) )
        {
            int i_read;

            /* Check which filedescriptor got the data */
            if( ! FD_ISSET( i_fd, &fdset ) )
            {
                continue;
            }

            /* Check that our buffer is large enough */
            if( i_buf_pos + ALLOC_ONCE > i_buf_size )
            {
                i_buf_size += ALLOC_ONCE;
                psz_buf = realloc( psz_buf, i_buf_size );
                if( psz_buf == NULL )
                {
                    ERROR( "WaitForData: failed reallocation on fd %i", i_fd );
                    close( i_fd );
                    return;
                }
            }

            /* Read data */
            i_read = read( i_fd, psz_buf + i_buf_pos, READ_ONCE );
            if( i_read < 0 )
            {
                ERROR( "WaitForData: error while reading on fd %i", i_fd );
                free( psz_buf );
                close( i_fd );
                return;
            }
            i_buf_pos += i_read;
        }
        else
        {
            /* No data: print what we got last time */
            if( i_buf_pos != 0 )
            {
                psz_buf[ i_buf_pos ] = '\0';
                INFO( "Descriptor %i got %i bytes: %s",
                      i_fd, i_buf_pos, psz_buf );

                if( ! strncmp( psz_buf, "\r", i_buf_pos - 1 ) )
                {
                    i_alive = 0;
                }

                i_buf_pos = 0;
            }
        }
    }
    while( i_alive );

    INFO( "Closing connection %i", i_fd );
    close( i_fd );
}

